(ns pong.core
  (:require [pong.menu :as menu]))

(enable-console-print!)

;; define your app data so that it doesn't get over-written on reload

(declare game-draw next-frame relaunch-ball game-loop)

(defonce keyboard (atom {}))

(def canvas (.getElementById js/document "game"))
(def context (.getContext canvas "2d"))
(def paddle-height 320)
(def paddle-width 40)
(def paddle-speed 1)
(def ball-radius 20)
(def ball-speed 0.75)


(defn window-height
  []
  (.-innerHeight js/window))

(defn window-width
  []
  (.-innerWidth js/window))

(defn unit-vector
  [v]
  (let [length (Math.sqrt (reduce + (map #(Math.pow % 2) v)))]
    (map #(/ % length) v)))

(defn contains-point?
  [rect x y]
  (and (> x (:x rect))
       (< x (+ (:x rect) (:w rect)))
       (> y (:y rect))
       (< y (+ (:y rect) (:h rect)))))

(defn collision?
  [ball paddle]
  (or (contains-point? paddle (- (:x ball) (:radius ball)) (:y ball))
      (contains-point? paddle (+ (:x ball) (:radius ball)) (:y ball))))

(defn apply-ball-velocity
  [{:keys [ball progress] :as state}]
  (let [velocity (map #(* % progress) (:velocity ball))]
    (assoc state :ball
           (assoc ball
                  :x (+ (first velocity) (:x ball))
                  :y (+ (second velocity) (:y ball))))))

(defn bounce-ball-off-top-wall
  [{{y :y radius :radius [vx vy] :velocity} :ball :as state}]
  (if (< (- y radius) 0)
    (as-> state state
        (assoc-in state [:ball :y] radius)
        (assoc-in state [:ball :velocity] [vx (- vy)]))
    state))

(defn bounce-ball-off-bottom-wall
  [{{y :y radius :radius [vx vy] :velocity} :ball :as state}]
  (if (> (+ y radius) (window-height))
    (as-> state state
        (assoc-in state [:ball :y] (- (window-height) radius))
        (assoc-in state [:ball :velocity] [vx (- vy)]))
    state))

(defn bounce-ball-off-paddle
  [{ball-y :y [dx] :velocity} {paddle-y :y paddle-h :h}]
  (unit-vector [(* (- (/ dx (Math.abs dx))) (window-width))
                (* (- (/ (- ball-y paddle-y) paddle-h) 0.5) (window-height))]))

(defn bounce-ball-off-left-paddle
  [{paddle :left-paddle ball :ball :as state}]
  (assoc-in state [:ball :velocity] (bounce-ball-off-paddle ball paddle)))

(defn bounce-ball-off-right-paddle
  [{paddle :right-paddle ball :ball :as state}]
  (assoc-in state [:ball :velocity] (bounce-ball-off-paddle ball paddle)))

(defn put-ball-infront-of-left-paddle
  [{{:keys [radius]} :ball {:keys [x w]} :left-paddle :as state}]
  (assoc-in state [:ball :x] (+ x w radius)))

(defn put-ball-infront-of-right-paddle
  [{{:keys [radius]} :ball {:keys [x w]} :right-paddle :as state}]
  (assoc-in state [:ball :x] (- x radius)))

(defn bounce-ball-off-left-paddle?
  [{:keys [ball left-paddle] :as state}]
  (if (collision? ball left-paddle)
    (-> state
        put-ball-infront-of-left-paddle
        bounce-ball-off-left-paddle)
    state))

(defn bounce-ball-off-right-paddle?
  [{:keys [ball right-paddle] :as state}]
  (if (collision? ball right-paddle)
    (-> state
        put-ball-infront-of-right-paddle
        bounce-ball-off-right-paddle)
    state))

(defn launch-ball-left
  [state]
  (assoc state :ball (relaunch-ball -1)))

(defn launch-ball-right
  [state]
  (assoc state :ball (relaunch-ball 1)))

(defn increase-right-player-score
  [state]
  (update-in state [:score :right] inc))

(defn increase-left-player-score
  [state]
  (update-in state [:score :left] inc))

(defn left-player-scored
  [{{:keys [x radius]} :ball :as state}]
  (if (> (- x radius) (window-width))
    (-> state
        launch-ball-right
        increase-left-player-score)
    state))

(defn right-player-scored
  [{{:keys [x radius]} :ball :as state}]
  (if (< (+ x radius) 0)
    (-> state
        launch-ball-left
        increase-right-player-score)
    state))

(defn apply-paddle-speed
  [paddle direction progress]
  (direction (:y paddle) (* paddle-speed progress)))

(defn handle-left-player-paddle-up-movement
  [{:keys [progress left-paddle] :as state}]
  (if (:KeyW @keyboard)
    (assoc-in state
              [:left-paddle :y]
              (apply-paddle-speed left-paddle - progress))
    state))

(defn handle-left-player-paddle-down-movement
  [{:keys [progress left-paddle] :as state}]
  (if (:KeyS @keyboard)
    (assoc-in state
              [:left-paddle :y]
              (apply-paddle-speed left-paddle + progress))
    state))

(defn keep-left-paddle-in-bounds
  [{:keys [left-paddle] :as state}]
  (assoc-in state
            [:left-paddle :y]
            (max 0 (min (- (window-height) (:h left-paddle))
                        (:y left-paddle)))))

(defn update-left-paddle
  [{:keys [progress left-paddle] :as state}]
  (-> state
      handle-left-player-paddle-down-movement
      handle-left-player-paddle-up-movement
      keep-left-paddle-in-bounds))

(defn stick-right-paddle-to-right-of-screen
  [{:keys [right-paddle] :as state}]
  (assoc-in state [:right-paddle :x] (- (window-width) (:w right-paddle))))

(defn handle-right-player-paddle-up-movement
  [{:keys [progress right-paddle] :as state}]
  (if (:ArrowUp @keyboard)
    (assoc-in state
              [:right-paddle :y]
              (apply-paddle-speed right-paddle - progress))
    state))

(defn handle-right-player-paddle-down-movement
  [{:keys [progress right-paddle] :as state}]
  (if (:ArrowDown @keyboard)
    (assoc-in state
              [:right-paddle :y]
              (apply-paddle-speed right-paddle + progress))
    state))

(defn player-update-right-paddle
  [state]
  (-> state
      handle-right-player-paddle-up-movement
      handle-right-player-paddle-down-movement))

(defn computer-update-right-paddle
  [{:keys [progress right-paddle ball] :as state}]
  (let [paddle-middle (+ (:y right-paddle) (/ (:h right-paddle) 2))
        speed (* paddle-speed progress)]
    (assoc-in state [:right-paddle :y]
              (cond
                (< (:y ball) (- paddle-middle speed))
                (max (- (:y right-paddle) speed) 0)
                (> (:y ball) (+ paddle-middle speed))
                (min (+ (:y right-paddle) speed)
                     (- (window-height) (:h right-paddle)))
                :else (:y right-paddle)))))

(defn draw-rect
  [{x :x y :y w :w h :h}]
  (.fillRect context x y w h))

(defn fill-circle
  [{x :x y :y radius :radius}]
  (.beginPath context)
  (.arc context x y radius 0 (* 2 Math.PI))
  (.fill context))

(defn draw-score
  [{:keys [left right]}]
  (set! (.-font context) "30px monospace")
  (.fillText context (str left " | " right) (/ (window-width) 2) 30))

(defn game-draw
  [{:keys [left-paddle right-paddle ball score] :as state}]
  (.clearRect context 0 0 (window-width) (window-height))
  (set! (.-fillStyle context) "#FFF")
  (draw-rect left-paddle)
  (draw-rect right-paddle)
  (fill-circle ball)
  (draw-score score)
  state)

(defn handle-return-to-menu
  [{:keys [keyboard] :as state}]
  (if (:Escape keyboard)
    (assoc state :on-loop menu/scene)
    state))

(defn game-loop
  [{:keys [update-right-paddle] :as state}]
  (-> state
      update-left-paddle
      update-right-paddle
      stick-right-paddle-to-right-of-screen
      apply-ball-velocity
      bounce-ball-off-top-wall
      bounce-ball-off-bottom-wall
      bounce-ball-off-left-paddle?
      bounce-ball-off-right-paddle?
      left-player-scored
      right-player-scored
      handle-return-to-menu
      game-draw))

(defn next-frame
  [last-render state]
  (.requestAnimationFrame js/window
                          #(next-frame % ((:on-loop state)
                                          (merge state
                                                 {:window-width (.-innerWidth js/window)
                                                  :window-height (.-innerHeight js/window)
                                                  :keyboard @keyboard
                                                  :progress (- % last-render)})))))

(defn vec*
  [v d]
  (map #(* % d) v))

(defn one-or-minus-one
  []
  (let [n (- (rand 2) 1)] (/ n (Math.abs n))))

(defn launch-ball [x y dx dy]
  {:x x
   :y y
   :radius ball-radius
   :velocity (vec* (unit-vector [dx dy]) ball-speed)})

(defn relaunch-ball
  [direction]
  (launch-ball (/ (window-width) 2)
               (rand (window-height))
               (* direction (/ (window-width) 2))
               (- (rand (window-height)) (/ (window-height) 2))))

(defn start-game
  [state update-right-paddle]
  (merge state
         {:left-paddle {:x 0.0
                        :y (/ (- (window-height) paddle-height) 2)
                        :w paddle-width
                        :h paddle-height}
          :right-paddle {:x (- (window-width) paddle-width)
                         :y (/ (- (window-height) paddle-height) 2)
                         :w paddle-width
                         :h paddle-height}
          :update-right-paddle update-right-paddle
          :ball (relaunch-ball (one-or-minus-one))
          :score {:left 0 :right 0}
          :on-loop game-loop}))

(defn start-one-player-game
  [state]
  (println "start one player game")
  (start-game state computer-update-right-paddle))

(defn start-two-players-game
  [state]
  (println "start two players game")
  (start-game state player-update-right-paddle))


(next-frame (.now js/performance)
            {:on-loop menu/scene
             :menu-item-selected 0
             :menu-items [{:txt "1 Player"
                           :action start-one-player-game}
                          {:txt "2 Players"
                           :action start-two-players-game}]
             :context context})

(defn set-canvas-size-to-window-size
  []
  (set! (.-width canvas) (.-innerWidth js/window))
  (set! (.-height canvas) (.-innerHeight js/window)))

(set-canvas-size-to-window-size)

(set! (.-onresize js/window)
      set-canvas-size-to-window-size)

;; (def update-game-state (partial swap! game-state assoc))
;; (def update-left-paddle-movement (partial update-game-state :left-paddle-movement))
;; (def update-right-paddle-movement (partial update-game-state :right-paddle-movement))

;; (swap! game-state assoc :on-loop menu-loop)
;; (fn [event]
;;                      (let [code (.-code event)]
;;                        (case code
;;                          "KeyW" (update-left-paddle-movement :up)
;;                          "KeyS" (update-left-paddle-movement :down)
;;                          "ArrowUp" (update-right-paddle-movement :up)
;;                          "ArrowDown" (update-right-paddle-movement :down)
;;                          "Escape" (swap! game-state assoc :on-loop menu-loop)
;;                          nil)))

;; (fn [event]
;;                      (let [code (.-code event)]
;;                        (case code
;;                          "KeyW" (and (= (:left-paddle-movement @game-state) :up)
;;                                      (update-left-paddle-movement nil))
;;                          "KeyS" (and (= (:left-paddle-movement @game-state) :down)
;;                                      (update-left-paddle-movement nil))
;;                          "ArrowUp" (and (= (:right-paddle-movement @game-state) :up)
;;                                      (update-right-paddle-movement nil))
;;                          "ArrowDown" (and (= (:right-paddle-movement @game-state) :down)
;;                                           (update-right-paddle-movement nil))
;;                          nil)))

(.addEventListener js/document "keydown" #(swap! keyboard assoc (keyword (.-code %)) %))
(.addEventListener js/document "keyup" #(swap! keyboard dissoc (keyword (.-code %))))

(defn on-js-reload []
  ;; optionally touch your app-state to force rerendering depending on
  ;; your application
  ;; (swap! app-state update-in [:__figwheel_counter] inc)
)
